use serde::{Deserialize, Serialize};
use sqlx::FromRow;

use crate::{get_pg, library::user::User};

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Playlist {
    // Unique identifier of the playlist
    pub id: uuid::Uuid,

    // Optional string representation of the playlist's ID
    #[sqlx(default)]
    pub id_str: Option<String>,

    // Owner of the playlist
    pub owner: String,

    // Human-readable title of the playlist
    pub title: String,

    // Level of visibility for this playlist (public, private, etc.)
    pub visibility: Visibility,

    // IDs of tracks that belong to this playlist
    pub tracks: Vec<uuid::Uuid>,
}

#[derive(Debug, Clone, Serialize, Deserialize, sqlx::Type)]
#[sqlx(type_name = "visibility", rename_all = "lowercase")]
pub enum Visibility {
    Private,
    Public,
}
impl Playlist {
    /// Creates a new playlist with the given owner, title, visibility, and tracks.
    ///
    /// # Arguments
    ///
    /// * `owner`: The user who owns the playlist.
    /// * `title`: The human-readable title of the playlist.
    /// * `visibility`: The level of visibility for this playlist (public, private, etc.).
    /// * `tracks`: A list of track IDs that belong to this playlist.
    pub async fn create(
        owner: User,
        title: &str,
        visibility: Visibility,
        tracks: &[uuid::Uuid],
    ) -> Option<Self> {
        sqlx::query_as("INSERT INTO playlist (owner, title, visibility, tracks) VALUES ($1, $2, $3, $4) RETURNING *")
            .bind(owner.username)
            .bind(title)
            .bind(visibility)
            .bind(tracks)
            .fetch_one(get_pg!())
            .await.ok()
    }

    /// Finds playlists where a given column matches a regular expression.
    ///
    /// # Arguments
    ///
    /// * `col`: The name of the column to search in (e.g., "title", "owner").
    /// * `query`: A string containing the regular expression pattern.
    ///
    /// # Returns
    ///
    /// A list of `Playlist` instances that match the query.
    pub async fn find_regex_col(col: &str, query: &str) -> Vec<Self> {
        sqlx::query_as(&format!("SELECT * FROM playlist WHERE {col} ~* $1"))
            // Bind the regular expression pattern as a string
            .bind(query)
            // Execute the query using the PostgreSQL connection pool and fetch all results
            .fetch_all(get_pg!())
            .await
            // Return the result as a list of `Playlist` instances (unwrap() is used here, but consider adding error handling)
            .unwrap()
    }

    /// Retrieves a playlist by its unique ID.
    ///
    /// # Arguments
    ///
    /// * `id`: The UUID of the playlist to retrieve.
    ///
    /// # Returns
    ///
    /// An optional `Playlist` instance if retrieval was successful; otherwise, `None`.
    pub async fn get(id: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM playlist WHERE id = $1")
            // Bind the UUID of the playlist to retrieve
            .bind(id)
            // Execute the query using the PostgreSQL connection pool and fetch an optional result
            .fetch_optional(get_pg!())
            // Return the result as an optional `Playlist` instance (unwrap() is used here, but consider adding error handling)
            .await
            .unwrap()
    }

    /// Retrieves all playlists owned by a given user.
    ///
    /// # Arguments
    ///
    /// * `u`: The user who owns the playlists to retrieve.
    ///
    /// # Returns
    ///
    /// A list of `Playlist` instances that belong to the given user.
    pub async fn of(u: &User) -> Vec<Self> {
        sqlx::query_as("SELECT * FROM playlist WHERE owner = $1")
            // Bind the username of the user who owns the playlists
            .bind(&u.username)
            // Execute the query using the PostgreSQL connection pool and fetch all results
            .fetch_all(get_pg!())
            // Return the result as a list of `Playlist` instances (unwrap() is used here, but consider adding error handling)
            .await
            .unwrap()
    }
}

impl Playlist {
    pub async fn api(&self) -> serde_json::Value {
        serde_json::json!({
            "id": if let Some(id) = &self.id_str { id.clone() } else { self.id.to_string() },
            "owner": self.owner,
            "visibility": serde_json::to_value(&self.visibility).unwrap(),
            "title": self.title,
            "tracks": self.tracks
        })
    }
}
