use serde::{Deserialize, Serialize};
use serde_json::json;
use sqlx::prelude::FromRow;
use std::{io::Read, str::FromStr};

use crate::{
    get_pg,
    library::album::Album,
    route::{to_uuid, ToAPI},
};

use super::{metadata::AudioMetadata, user::User};

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Track {
    /// The unique identifier of the track
    pub id: uuid::Uuid,

    // The file path where the track is stored
    pub path: String,

    /// The title of the track
    pub title: String,

    /// The date when the track was added to the library
    pub date_added: chrono::DateTime<chrono::Utc>,

    /// Optional associated album
    pub album: Option<uuid::Uuid>,

    /// Optional associated artist
    pub artist: Option<uuid::Uuid>,

    /// Optional metadata for the track
    pub meta: Option<serde_json::Value>,
}

impl Track {
    /// Create a new track.
    ///
    /// # Arguments
    ///
    /// * `data` - The data to create the track with.
    pub async fn create(data: &serde_json::Map<String, serde_json::Value>) {
        sqlx::query(
            "INSERT INTO track (path, title, meta, album, artist) VALUES ($1, $2, $3, $4, $5)",
        )
        .bind(data.get("path").unwrap().as_str().unwrap().to_string())
        .bind(data.get("title").unwrap().as_str().unwrap().to_string())
        .bind(data.get("meta"))
        .bind(
            data.get("album")
                .map(|x| to_uuid(x.as_str().unwrap()).unwrap()),
        )
        .bind(
            data.get("artist")
                .map(|x| to_uuid(x.as_str().unwrap()).unwrap()),
        )
        .execute(get_pg!())
        .await
        .unwrap();
    }

    /// Find a track by path.
    ///
    /// # Arguments
    ///
    /// * `path` - The path of the track to find.
    ///
    /// # Returns
    ///
    /// The found track, or None if no track is found.
    pub async fn of_path(path: &str) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE path = $1")
            .bind(path)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    /// Find all recently added tracks.
    ///
    /// # Returns
    ///
    /// A vector of recently added tracks.
    pub async fn find_recently_added() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM track ORDER BY date_added DESC LIMIT 90")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Find a track by ID.
    ///
    /// # Arguments
    ///
    /// * `id` - The ID of the track to find.
    ///
    /// # Returns
    ///
    /// The found track, or None if no track is found.
    pub async fn get(id: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE id = $1")
            .bind(id)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    /// Update a track.
    ///
    /// # Arguments
    ///
    /// * `update_set` - The update set to apply to the track.
    ///
    /// # Returns
    ///
    /// Some(()) if the update was successful, or None if not.
    pub async fn update(&self, update_set: &serde_json::Value) -> Option<()> {
        let map = update_set.as_object()?;
        let artist = map
            .get("artist")
            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap());
        let album = map
            .get("album")
            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap());
        let title = map.get("title")?.as_str()?;
        let meta = map.get("meta");

        sqlx::query(
            "UPDATE track SET artist = $1, album = $2, title = $3, meta = $4 WHERE id = $5",
        )
        .bind(artist)
        .bind(album)
        .bind(title)
        .bind(meta)
        .bind(self.id)
        .execute(get_pg!())
        .await
        .unwrap();

        Some(())
    }

    /// Find all tracks.
    ///
    /// # Returns
    ///
    /// A vector of all tracks.
    pub async fn find_all() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM track")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Remove a track.
    ///
    /// # Arguments
    ///
    /// * `id` - The ID of the track to remove.
    pub async fn remove(&self) {
        sqlx::query("DELETE FROM track WHERE id = $1")
            .bind(self.id)
            .fetch_one(get_pg!())
            .await
            .unwrap();
    }

    /// Find tracks by album.
    ///
    /// # Arguments
    ///
    /// * `album` - The album to find tracks for.
    ///
    /// # Returns
    ///
    /// A vector of tracks with the specified album.
    pub async fn find_of_album(album: &uuid::Uuid) -> Vec<Self> {
        sqlx::query_as("SELECT * FROM track WHERE album = $1")
            .bind(album)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Retrieves a list of tracks that are the latest listened to for a given user.
    ///
    /// This function fetches the latest events associated with the user and then filters them by track ID.
    /// It uses these IDs to query the database for all tracks with matching IDs.
    ///
    /// # Arguments
    ///
    /// * `u`: The user object whose latest tracks are being retrieved.
    ///
    /// # Returns
    ///
    /// A vector of `Track` objects representing the latest tracks for the given user.
    pub async fn get_latest_of_user(u: &User) -> Vec<Self> {
        let ids: Vec<(uuid::Uuid, chrono::DateTime<chrono::Utc>)> = sqlx::query_as(
            "SELECT DISTINCT(track), time FROM events WHERE \"user\" = $1 ORDER BY time DESC LIMIT 300",
        )
        .bind(&u.username)
        .fetch_all(get_pg!())
        .await
        .unwrap();

        let mut tracks: Vec<_> = Vec::with_capacity(300);

        for track in ids {
            tracks.push(Track::get(&track.0).await.unwrap());
        }

        tracks
    }

    /// Transcodes audio to OPUS format with the specified bitrate.
    ///
    /// # Arguments
    ///
    /// * `bitrate`: The desired bitrate for the transcoded audio.
    ///
    /// # Returns
    ///
    /// A `String` representing the path to the transceded audio file, or `None` if the transcoding failed.
    pub fn get_opus(&self, bitrate: u32) -> Option<String> {
        self.transcode("libopus", bitrate, "opus")
    }

    /// Transcodes audio to AAC format with the specified bitrate.
    ///
    /// # Arguments
    ///
    /// * `bitrate`: The desired bitrate for the transcoded audio.
    ///
    /// # Returns
    ///
    /// A `String` representing the path to the transceded audio file, or `None` if the transcoding failed.
    pub fn get_aac(&self, bitrate: u32) -> Option<String> {
        self.transcode("aac", bitrate, "aac")
    }

    /// Transcodes audio from its original format to the specified codec and extension.
    ///
    /// This function creates a new directory for the transceded audio if it does not exist,
    /// and then uses `ffmpeg` to perform the transcoding operation.
    ///
    /// # Arguments
    ///
    /// * `codec`: The desired codec for the transceded audio (e.g. "libopus", "aac").
    /// * `bitrate`: The desired bitrate for the transceded audio.
    /// * `ext`: The desired extension for the transceded audio file (e.g. "opus", "aac").
    ///
    /// # Returns
    ///
    /// A `String` representing the path to the transceded audio file, or `None` if the transcoding failed.
    pub fn transcode(&self, codec: &str, bitrate: u32, ext: &str) -> Option<String> {
        let transcoded = format!("./data/transcode/{codec}/{bitrate}/{}.{ext}", self.id);

        if std::path::Path::new(&transcoded).exists() {
            return Some(transcoded);
        }

        log::info!("Transcoding {} to {} {}", self.id, codec, bitrate);

        std::fs::create_dir_all(format!("./data/transcode/{codec}/{bitrate}")).unwrap();

        let out = std::process::Command::new("ffmpeg")
            .arg("-i")
            .arg(&self.path)
            .arg("-c:a")
            .arg(codec)
            .arg("-b:a")
            .arg(format!("{bitrate}k"))
            .arg(&transcoded)
            .output()
            .unwrap();

        if !out.status.success() {
            return None;
        }

        Some(transcoded)
    }

    /// Finds tracks in the database that match a regular expression pattern.
    ///
    /// # Arguments
    ///
    /// * `col`: The name of the column to filter on.
    /// * `query`: The regular expression pattern to match against.
    ///
    /// # Returns
    ///
    /// A vector of `Track` objects representing the matching tracks.
    pub async fn find_regex_col(col: &str, query: &str) -> Vec<Self> {
        sqlx::query_as(&format!("SELECT * FROM track WHERE {col} ~* $1"))
            .bind(query)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Retrieves a list of tracks with no album or artist associated.
    ///
    /// # Returns
    ///
    /// A vector of `Track` objects representing the orphaned tracks.
    pub async fn get_orphans() -> Vec<Track> {
        sqlx::query_as("SELECT * FROM track WHERE artist IS NULL AND album IS NULL")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Finds the first track associated with a given artist.
    pub async fn find_first_of_artist(artist: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE artist = $1 LIMIT 1")
            .bind(artist)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get_lyrics(&self) -> Option<serde_json::Value> {
        let base_path = std::path::Path::new(&self.path).with_extension("");
        let extensions = vec!["lrc", "txt"];

        for ext in &extensions {
            let candidate = base_path.with_extension(ext);
            if candidate.exists() {
                if let Ok(mut file) = std::fs::File::open(&candidate) {
                    let mut contents = String::new();
                    if file.read_to_string(&mut contents).is_ok() {
                        // Determine if it's timed lyrics or plain text
                        let lyrics_type = if ext == &"lrc" { "timed" } else { "plain" };

                        return Some(json!({
                            "lyrics": contents,
                            "source": "file",
                            "type": lyrics_type
                        }));
                    }
                }
            }
        }

        if let Some(lyrics) = self.meta.as_ref()?.get("lyrics").and_then(|l| l.as_str()) {
            // Check if lyrics seem to be timed or plain text
            let lyrics_type = if lyrics.contains('[') && lyrics.contains(']') {
                "timed" // A rough check for LRC-style brackets indicating timing
            } else {
                "plain"
            };

            return Some(json!({
                "lyrics": lyrics,
                "type": lyrics_type
            }));
        }

        // No lyrics found
        None
    }

    /// Finds the first track of a given album.
    pub async fn find_first_of_album(album: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE album = $1 LIMIT 1")
            .bind(album)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }
}

impl ToAPI for Track {
    async fn api(&self) -> serde_json::Value {
        // todo : fix
        let (cover, album_title, album_id) = if let Some(album_ref) = &self.album {
            let album = Album::get(album_ref).await.unwrap();

            (album.get_cover().await.is_some(), album.title, album.id)
        } else {
            (false, String::new(), uuid::Uuid::nil())
        };

        let artist_title = if let Some(artist_ref) = &self.artist {
            let res: (String,) = sqlx::query_as("SELECT name FROM artist WHERE id = $1")
                .bind(artist_ref)
                .fetch_one(get_pg!())
                .await
                .unwrap();
            Some(res.0)
        } else {
            None
        };

        json!({
            "id": self.id,
            "title": self.title,
            "track_number": self.meta.as_ref().map(|x| AudioMetadata(x.clone()).track_number()),
            "meta": serde_json::to_value(&self.meta).unwrap(),
            "album_id": self.album,
            "album": album_title,
            "cover": if cover {
                Some(format!("/album/{album_id}/cover"))
            } else {
                None
            },
            "artist_id": self.artist,
            "artist": artist_title
        })
    }
}
