use std::cmp::Ordering;

use super::api_error;
use super::to_uuid;
use super::FallibleApiResponse;
use rocket::fs::NamedFile;
use rocket::{get, State};
use serde_json::json;

use crate::cache::RouteCache;
use crate::library::album::Album;
use crate::library::track::Track;
use crate::library::Libary;
use crate::route::vec_to_api;
use crate::route::ToAPI;
use crate::use_api_cache;

#[get("/artist/<artist_id>/albums")]
pub async fn albums_route(artist_id: &str, lib: &State<Libary>) -> FallibleApiResponse {
    let albums = lib.get_albums_by_artist(&to_uuid(artist_id)?).await;
    let singles = lib.get_singles_by_artist(&to_uuid(artist_id)?).await;

    Ok(json!({
        "artist": artist_id,
        "albums": vec_to_api(&albums).await,
        "singles": vec_to_api(&singles).await
    }))
}

fn sort_by_tracknumber(a: &serde_json::Value, b: &serde_json::Value) -> Ordering {
    let order = (|| {
        Some(
            a.get("track_number")?
                .as_i64()?
                .cmp(&b.get("track_number")?.as_i64()?),
        )
    })();
    order.unwrap_or(Ordering::Equal)
}

#[get("/album/<album_id>/cover")]
pub async fn album_cover_route(
    album_id: &str,
    lib: &State<Libary>,
    cache: &State<RouteCache>,
) -> Option<NamedFile> {
    NamedFile::open(
        cache
            .get_option("album_cover_route", album_id, || async {
                let album = lib.get_album_by_id(&to_uuid(album_id).unwrap()).await?;
                album.get_cover().await
            })
            .await?,
    )
    .await
    .ok()
}

#[get("/albums/latest")]
pub async fn latest_albums_route(cache: &State<RouteCache>) -> FallibleApiResponse {
    // todo : fix

    use_api_cache!("albums", "latest", cache);

    let albums = Album::find_all().await;

    let mut albums_tracks = vec![];

    for album in &albums {
        albums_tracks.push(Track::find_first_of_album(&album.id).await.unwrap());
    }

    let mut joined: Vec<(_, _)> = albums.into_iter().zip(albums_tracks).collect();

    joined.sort_by(|a, b| a.1.date_added.cmp(&b.1.date_added));

    joined.reverse();

    let (albums, _): (Vec<_>, Vec<_>) = joined.into_iter().unzip();

    let albums = vec_to_api(&albums).await;

    cache
        .insert("albums", "latest", serde_json::to_string(&albums).unwrap())
        .await;

    Ok(json!(albums))
}

#[get("/album/<album_id>")]
pub async fn album_route(
    album_id: &str,
    lib: &State<Libary>,
    cache: &State<RouteCache>,
) -> FallibleApiResponse {
    use_api_cache!("album_route", album_id, cache);

    let album = lib
        .get_album_by_id(&to_uuid(album_id)?)
        .await
        .ok_or_else(|| api_error("No album with that ID found"))?;

    let tracks = Album::get_tracks_of_album(&album.id).await;

    let mut tracks = vec_to_api(&tracks).await;

    tracks.sort_by(sort_by_tracknumber);

    let mut album = album.api().await;
    album
        .as_object_mut()
        .unwrap()
        .insert("tracks".into(), tracks.into());

    cache
        .insert(
            "album_route",
            album_id,
            serde_json::to_string(&album).unwrap(),
        )
        .await;

    Ok(album)
}
