use std::str::FromStr;

use super::api_error;
use super::no_uuid_error;
use super::to_uuid;
use super::FallibleApiResponse;
use super::ToAPI;
use crate::library::user::User;
use fs::NamedFile;
use rocket::{fs, get, State};
use serde_json::json;

use crate::check_admin;
use crate::library::Libary;

#[get("/track/<track_id>")]
pub async fn track_route(track_id: &str, lib: &State<Libary>) -> FallibleApiResponse {
    Ok(lib
        .get_track_by_id(&to_uuid(track_id)?)
        .await
        .ok_or_else(|| api_error("No track with that ID found"))?
        .api()
        .await)
}

#[get("/track/<track_id>/reload")]
pub async fn track_reload_meta_route(
    track_id: &str,
    lib: &State<Libary>,
    u: User,
) -> FallibleApiResponse {
    check_admin!(u);
    lib.reload_metadata(&to_uuid(track_id)?)
        .await
        .map_err(|_| api_error("Error reloading metadata"))?;
    Ok(json!({"ok": 1}))
}

#[get("/track/<track_id>/audio")]
pub async fn track_audio_route(track_id: &str, lib: &State<Libary>) -> Option<NamedFile> {
    let track = lib
        .get_track_by_id(&uuid::Uuid::from_str(track_id).ok()?)
        .await?;
    NamedFile::open(std::path::Path::new(&track.path))
        .await
        .ok()
}

#[get("/track/<track_id>/audio/opus128")]
pub async fn track_audio_opus128_route(track_id: &str, lib: &State<Libary>) -> Option<NamedFile> {
    let track = lib
        .get_track_by_id(&uuid::Uuid::from_str(track_id).ok()?)
        .await?;
    NamedFile::open(track.get_opus(128)?).await.ok()
}

#[get("/track/<track_id>/audio/aac128")]
pub async fn track_audio_aac128_route(track_id: &str, lib: &State<Libary>) -> Option<NamedFile> {
    let track = lib
        .get_track_by_id(&uuid::Uuid::from_str(track_id).ok()?)
        .await?;
    NamedFile::open(track.get_aac(128)?).await.ok()
}

#[get("/track/<track_id>/lyrics")]
pub async fn track_lyrics_route(track_id: &str, lib: &State<Libary>) -> FallibleApiResponse {
    let track = lib
        .get_track_by_id(&uuid::Uuid::from_str(track_id).map_err(|_| no_uuid_error())?)
        .await
        .ok_or_else(|| api_error("No such track"))?;

    if let Some(lyrics) = track.get_lyrics().await {
        return Ok(lyrics);
    }

    Ok(json!({}))
}
